<?php
namespace Teknisa\Libs\Service;

use Teknisa\Libs\Util\InstanceProvider;
use Teknisa\Libs\Util\Repositories\Migration as MigrationRepository;
use Teknisa\Libs\Util\Utilities;

class Migration {
    /** @var MigrationRepository $migrationRepository */
    private $migrationRepository;

    /**
     * @param MigrationRepository $migrationRepository
     */
    public function __construct(MigrationRepository $migrationRepository) {
        $this->migrationRepository = $migrationRepository;
    }

    /**
     * @param $nrOrg
     * @param $repositoriesInfo
     * @return array
     */
    public function getPendentMigrations($nrOrg, $repositoriesInfo) {
        try {
            $migrationInfo = array();
            $productPath   = Utilities::getProductBasePath();

            $customEntityManager = InstanceProvider::getConnectionCustom()->updateVpdConnection($nrOrg);
            if(is_string($repositoriesInfo)) {
                $repositoriesInfo = json_decode($repositoriesInfo, true);
            }

            foreach ($repositoriesInfo as $repository) {
                $migDescription  = substr($repository['URL'], -50);
                $scriptsPath     = $productPath . DIRECTORY_SEPARATOR . $repository['PATH'] . DIRECTORY_SEPARATOR . $repository['MIG_RELATIVE_PATH'];

                if(isset($migrationInfo[$migDescription])) {
                    $lastFile  = $migrationInfo[$migDescription]['MIG_LAST_FILE_NUMBER'];
                    $pendFiles = $migrationInfo[$migDescription]['MIG_PENDENT_FILES'];
                } else {
                    $versionObj = $this->migrationRepository->getLastAppliedMigration($migDescription, $nrOrg, $customEntityManager);
                    $lastFile   = !empty($versionObj) ? $versionObj[0]['NRVEROBJBD']   : null;
                    $pendFiles  = !empty($versionObj) ? @json_decode($versionObj[0]['DSARQPENDMIG'], true) : array();
                    $pendFiles  = is_array($pendFiles) ? $pendFiles : array();
                }

                $higherFileNumber = $this->getHigherFileNumber($scriptsPath);

                if(isset($migrationInfo[$migDescription])) {
                    if($lastFile && $higherFileNumber > $migrationInfo[$migDescription]['MIG_HIGHER_FILE_NUMBER']) {
                        $migrationInfo[$migDescription]['MIG_HIGHER_FILE_NUMBER'] = $higherFileNumber;
                    }
                } else {
                    $migrationInfo[$migDescription] = array (
                        'REPOSITORY_ID'          => $repository['ID'],
                        'REPOSITORY_URL'         => $repository['URL'],
                        'REPOSITORY_NAME'        => $repository['NAME'],
                        'REPOSITORY_PATH'        => $repository['PATH'],
                        'REPOSITORY_REL_PATH'    => $repository['MIG_RELATIVE_PATH'],
                        'MIG_DESCRIPTION'        => $migDescription,
                        'MIG_LAST_FILE_NUMBER'   => $lastFile,
                        'MIG_HIGHER_FILE_NUMBER' => $higherFileNumber,
                        'MIG_PENDENT_FILES'      => $pendFiles
                    );
                }
            }

            return array_values($migrationInfo);
        } catch (\Exception $e) {
            return $this->errorTreatmentOnSQLExecution($e);
        }
    }

    /**
     * @param $scriptsPath
     * @return int|string
     */
    private function getHigherFileNumber($scriptsPath) {
        $files = array_diff(scandir($scriptsPath), array('..', '.'));
        $files = $this->orderDirectory($files);
        $higherFileNumber = 0;

        foreach($files as $file) {
            if(strpos($file, 'down') === false) {
                $delimiter  = strpos($file, '_') ? '_' : '.';
                $fileNumber = explode($delimiter, $file)[0];
                if (is_numeric($fileNumber)) {
                    $higherFileNumber = $fileNumber;
                    break;
                }
            }
        }

        return intval($higherFileNumber);
    }

    /**
     * @param $fileNumber
     * @param $previousFileNumber
     * @param $pendFiles
     * @return array
     */
    public function checkPendentFiles($fileNumber, $previousFileNumber, $pendFiles) {
        $auxFileNumber = $fileNumber;
        if($previousFileNumber) {
            while ($auxFileNumber - $previousFileNumber > 1) {
                --$auxFileNumber;
                if(!in_array($auxFileNumber, $pendFiles)) {
                    $pendFiles[] = $auxFileNumber;
                }
            }
        }
        if(in_array($fileNumber, $pendFiles)) {
            $key = array_search($fileNumber, $pendFiles);
            unset($pendFiles[$key]);
        }
        return $pendFiles;
    }

    /**
     * @param $files
     * @return mixed
     */
    private function orderDirectory($files) {
        usort($files, function($currentFile, $nextFile) {
            if(strpos($currentFile, 'down') === false) {
                $delimiter         = strpos($currentFile, '_') ? '_' : '.';
                $currentFileNumber = intval(explode($delimiter, $currentFile)[0]);
                $nextFileNumber    = intval(explode($delimiter, $nextFile)[0]);
                return $currentFileNumber > $nextFileNumber ? -1 : 1;
            } else {
                return 0;
            }
        });
        return $files;
    }

    /**
     * @param $fileNumber
     * @param $higherFileNumber
     * @param $smallFileNumber
     * @param $setSmall
     */
    private function setHigherAndSmallFileNumber($fileNumber, &$higherFileNumber, &$smallFileNumber, $setSmall) {
        if($fileNumber > $higherFileNumber) {
            $higherFileNumber = $fileNumber;
        }

        if((!$smallFileNumber || $fileNumber < $smallFileNumber) && $setSmall) {
            $smallFileNumber = $fileNumber;
        }
    }

    /**
     * @param $script
     * @param $nrOrg
     * @return array
     * @throws \Exception
     */
    public function applyMigration($script, $nrOrg) {
        try {
            $customEntityManager = InstanceProvider::getConnectionCustom()->updateVpdConnection($nrOrg);
            $result = $this->migrationRepository->executeScript($script, $customEntityManager);
            return $this->errorTreatmentOnSQLExecution($result);
        } catch (\Exception $e) {
            return $this->errorTreatmentOnSQLExecution($e);
        }
    }

    /**
     * @param $nmObject
     * @param $fileName
     * @param $oldFile
     * @param $nrOrg
     * @param $cdOperator
     * @param $pendentFiles
     * @return array
     */
    public function saveLastSuccessMigration($nmObject, $fileName, $oldFile, $nrOrg, $cdOperator, $pendentFiles = '[]') {
        try {
            $customEntityManager = InstanceProvider::getConnectionCustom()->updateVpdConnection($nrOrg);
            if($oldFile) {
                $result = $this->migrationRepository->updateLastMigration($nmObject, $fileName, $nrOrg, $cdOperator, $customEntityManager, $pendentFiles);
            } else {
                $result = $this->migrationRepository->insertLastMigration($nmObject, $fileName, $nrOrg, $cdOperator, $customEntityManager, $pendentFiles);
            }
            return $this->errorTreatmentOnSQLExecution($result);
        } catch (\Exception $e) {
            return $this->errorTreatmentOnSQLExecution($e);
        }
    }

    /**
     * @param $result
     * @return array
     */
    private function errorTreatmentOnSQLExecution($result) {
        $error = $rowCount = null;

        if($result instanceof \Exception) {
            /** @var \Exception $result */
            $error = $result->getMessage();
        } else {
            /** @var \Doctrine\DBAL\Driver\Statement $result */
            $errorInfo = $result->errorInfo();
            if(isset($errorInfo['message']) && !empty($errorInfo['message'])) {
                $error = $errorInfo['message'];
            }
            $rowCount = $result->rowCount();
        }

        if(strpos($error, 'ORA-') !== false) {
            $error = 'ORA-' . explode('ORA-', $error)[1];
        }

        return array(
            'ROW_COUNT' => $rowCount,
            'ERROR'     => $error
        );
    }
}